#!/usr/bin/env python
# @Author: carlosgilgonzalez
# @Date:   2019-07-23T00:24:19+01:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-07-25T18:28:23+01:00


import utime


def tzero():
    return utime.ticks_us()


def tdiff(t0):
    return utime.ticks_diff(utime.ticks_us(), t0)


def result(module_name, delta, cmd=False):
    if cmd:
        print('Command {} Time = {:6.3f}ms'.format(module_name, delta/1000))
    else:
        print('Script {} Time = {:6.3f}ms'.format(module_name, delta/1000))



#!/usr/bin/env python
# @Author: carlosgilgonzalez
# @Date:   2019-07-09T19:42:05+01:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-11-17T04:28:54+00:00

import usocket as socket
import os
import gc
import time

buffer_chunk = bytearray(2000)


def connect_SOC(host, port):
    cli_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    cli_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    soc_addr = socket.getaddrinfo(host, port)[0][-1]
    cli_soc.connect(soc_addr)
    return cli_soc


def print_sizefile(path, file_name, tabs=0, rtl=False):
    files = [filename for filename in os.listdir(
        path) if filename == file_name]
    for file in files:
        try:
            stats = os.stat(path + "/" + file)
        except Exception as e:
            try:
                stats = os.stat(file)
            except Exception as e:
                print('File not found.')
        filesize = stats[6]
        isdir = stats[0] & 0x4000

        _kB = 1024
        if filesize < _kB:
            sizestr = str(filesize) + " by"
        elif filesize < _kB**2:
            sizestr = "%0.1f KB" % (filesize / _kB)
        elif filesize < _kB**3:
            sizestr = "%0.1f MB" % (filesize / _kB**2)
        else:
            sizestr = "%0.1f GB" % (filesize / _kB**3)

        prettyprintname = ""
        for _ in range(tabs):
            prettyprintname += "   "
        prettyprintname += file
        if isdir:
            prettyprintname += "/"
        print('{0:<40} Size: {1:>10}'.format(prettyprintname, sizestr))
    if rtl:
        return filesize


def read_sd_file_sync(file_in_sd, soc):  # for files wrote with new line separator(\n)
    with open(file_in_sd, 'r') as log:
        while True:
            try:
                line = log.readline()
                if line is not '':
                    # in python use 'i'
                    soc.sendall(line[:-1])
                else:
                    break
            except Exception as e:
                print(e)
                pass


# for files wrote with new line separator(\n)
def read_sd_file_sync_raw(file_in_sd, soc, buff=buffer_chunk):
    # final_file = b''
    if os.getcwd() == '/sd':
        file_in_sd = file_in_sd.split('/')[-1]
    with open(file_in_sd, 'rb') as log:
        while True:
            try:
                buff[:] = log.read(2000)  # 2 KB
                # print(len(chunk))
                if buff != b'':
                    # in python use 'i'
                    soc.sendall(buff)
                    # final_file += chunk
                else:
                    # print('END OF FILE')
                    # soc.sendall(b'EOF\x8a\xb1\x1a\xcb\x11')
                    # soc.close()
                    gc.collect()
                    break
            except Exception as e:
                print(e)
                pass


def sync_to_filesys(filetoget, host, port, buff=buffer_chunk):
    # final_file = b''
    soc = connect_SOC(host, port)
    soc.settimeout(2)
    with open(filetoget, 'wb') as log:
        pass
    while True:
        try:
            chunk = soc.recv(2000)  # 2 KB
            if chunk != b'':
                with open(filetoget, 'ab') as log:
                    log.write(chunk)
            else:
                print('END OF FILE')
                break
        except Exception as e:
            if e == KeyboardInterrupt:
                break
    gc.collect()
    # soc.close()


def w_stream_writer(host, port, chunk=20000, total=10000000):
    soc = connect_SOC(host, port)
    print('Connected!')
    print('Streaming...\n')
    n_times_chunk = int(total/chunk)
    buff = os.urandom(chunk)
    try:
        for _ in range(n_times_chunk):
            soc.sendall(buff)  # 20 kB
    except Exception as e:
        print(e)
        if e == KeyboardInterrupt:
            print(e)

    time.sleep(0.1)
    print('\nupydevice Done!')
    # soc.close()
    gc.collect()



# !/usr/bin/env python3
# @Author: carlosgilgonzalez
# @Date:   2019-11-10T19:52:25+00:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-11-14T22:39:25+00:00


from machine import Pin, PWM, Timer
import time


class NOTIFIER:
    def __init__(self, buzz_pin, led_pin, fq=4000, on_time=150, n_times=2,
                 off_time=50, timer=None, period=5000):

        self.led = Pin(led_pin, Pin.OUT)
        self.fq = fq
        self.duty = 512
        self.buzz = PWM(Pin(buzz_pin), freq=self.fq, duty=self.duty)
        self.buzz.deinit()
        self.on_time = on_time
        self.n_times = n_times
        self.off_time = off_time
        self.period = period
        self.irq_busy = False
        self.tim = timer
        self.blink = True
        self.use_dict = {'buzz': self.buzzer_call, 'led': self.blink_call}
        if timer is not None:
            self.tim = Timer(timer)

    def buzz_beep(self, beep_on_time, n_times, beep_off_time, fq, led=True):
        self.buzz.freq(fq)
        if led:
            for i in range(n_times):
                self.buzz.init()
                self.led.on()
                time.sleep_ms(beep_on_time)
                self.buzz.deinit()
                self.led.off()
                time.sleep_ms(beep_off_time)
        else:
            for i in range(n_times):
                self.buzz.init()
                time.sleep_ms(beep_on_time)
                self.buzz.deinit()
                time.sleep_ms(beep_off_time)

    def led_blink(self, led_on_time, n_times, led_off_time):
        for i in range(n_times):
            self.led.on()
            time.sleep_ms(led_on_time)
            self.led.off()
            time.sleep_ms(led_off_time)

    def buzzer_call(self, x):
        if self.irq_busy:
            return
        else:
            self.irq_busy = True
            self.buzz_beep(self.on_time, self.n_times, self.off_time,
                           self.fq, self.blink)
            self.irq_busy = False

    def blink_call(self, x):
        if self.irq_busy:
            return
        else:
            self.irq_busy = True
            self.led_blink(self.on_time, self.n_times, self.off_time)
            self.irq_busy = False

    def notify(self, use='buzz', mode='SHOT', timeout=5000, on_init=None):
        self.irq_busy = False
        if on_init is not None:
            on_init()

        notify_call = self.use_dict[use]

        if mode == 'SHOT':
            self.tim.init(period=timeout, mode=Timer.ONE_SHOT,
                          callback=notify_call)
        elif mode == 'PERIODIC':
            self.tim.init(period=timeout, mode=Timer.PERIODIC,
                          callback=notify_call)

    def stop_notify(self):
        self.tim.deinit()
        self.irq_busy = False



import sys
from micropython import const
import time


CRITICAL = const(50)
ERROR = const(40)
WARNING = const(30)
INFO = const(20)
DEBUG = const(10)
NOTSET = const(0)

_level_dict = {
    CRITICAL: "CRIT",
    ERROR: "ERROR",
    WARNING: "WARN",
    INFO: "INFO",
    DEBUG: "DEBUG",
}

_format_dict = {"LVL_MSG": "[{}] [{}] ", "TIME_LVL_MSG": "{} [{}] [{}] "}

_stream = sys.stderr
_level = INFO
_loggers = {}
_filename = 'error.log'
_format = _format_dict["LVL_MSG"]
_asciitime = False
_sdlog = False


class Logger:
    def __init__(self, name, l_level=_level, log_to_file=False, logfile=_filename, f_time=False, l_format=_format):
        self.name = name
        self.level = l_level
        self.logfile = logfile
        self.log_to_file = log_to_file
        self.l_format = l_format
        self.log_message = None
        self.log_message_info = None
        self.f_time = f_time
        self._t_now = None
        self.date_time = None
        self._n = None
        self._dt_list = [0, 1, 2, 3, 4, 5]

    def _level_str(self, level):
        lev = _level_dict.get(level)
        if lev is not None:
            return lev
        return "LVL: {}".format(level)

    def setLevel(self, level):
        self.level = level

    def _dt_format(self, number):
        self._n = str(number)
        if len(self._n) == 1:
            self._n = "0{}".format(self._n)
            return self._n
        else:
            return self._n

    def _ft_datetime(self, t_now):
        return([self._dt_format(t_now[i]) for i in self._dt_list])

    def get_datetime(self):
        self.t_now = time.localtime()
        self.date_time = '{}-{}-{} {}:{}:{}'.format(*self._ft_datetime(self.t_now))
        return self.date_time

    def isEnabledFor(self, level):
        return level >= (self.level or _level)

    def file_log_msg(self, msg):
        with open(self.logfile, 'ab') as flog:
            flog.write(msg)
            flog.write('\n')

    def file_log_exception(self, ex):
        with open(self.logfile, 'ab') as flog:
            sys.print_exception(ex, flog)
            flog.write('\n')

    def log(self, level, msg, *args):
        if level >= (self.level or _level):
            if self.f_time:
                self.get_datetime()
                self.log_message_info = self.l_format.format(self.date_time,
                                                             self.name,
                                                             self._level_str(level),
                                                             )
            else:
                self.log_message_info = self.l_format.format(self.name,
                                                             self._level_str(level),
                                                             )
            _stream.write(self.log_message_info)
            if not args:
                print(' '.join([msg]), file=_stream)
                if self.log_to_file:
                    self.file_log_msg(''.join([self.log_message_info, msg]))
            else:
                print(' '.join([msg % args]), file=_stream)
                if self.log_to_file:
                    self.file_log_msg(''.join([self.log_message_info, msg % args]))

    def debug(self, msg, *args):
        self.log(DEBUG, msg, *args)

    def info(self, msg, *args):
        self.log(INFO, msg, *args)

    def warning(self, msg, *args):
        self.log(WARNING, msg, *args)

    def error(self, msg, *args):
        self.log(ERROR, msg, *args)

    def critical(self, msg, *args):
        self.log(CRITICAL, msg, *args)

    def exception(self, e, msg, *args):
        self.log(ERROR, msg, *args)
        sys.print_exception(e, _stream)
        if self.log_to_file:
            self.file_log_exception(e)


def getLogger(name, log_to_file=False):
    global _level, _stream, _filename, _format, _format_dict, _asciitime, _sdlog
    if name in _loggers:
        return _loggers[name]
    ulogger = Logger(name, l_level=_level,
                     log_to_file=log_to_file, logfile=_filename,
                     f_time=_asciitime, l_format=_format)
    _loggers[name] = ulogger
    return ulogger


def info(msg, *args):
    getLogger(None).info(msg, *args)


def debug(msg, *args):
    getLogger(None).debug(msg, *args)


def basicConfig(level=INFO, filename=_filename, stream=None, format=None, sd=False):
    global _level, _stream, _filename, _format, _format_dict, _asciitime, _sdlog
    _level = level
    if stream:
        _stream = stream
    if filename is not None:
        _filename = filename
    if format is not None:
        if format not in _format_dict.keys():
            print('Not supported format; Supported format are : {}'.format(_format_dict.keys()))
        else:
            _format = _format_dict[format]
            if format == 'TIME_LVL_MSG':
                _asciitime = True
    if sd:
        _sdlog = True
        if filename is not None:
            _filename = "/sd/{}".format(filename)



#!/usr/bin/env python

from ubinascii import hexlify, unhexlify
from machine import unique_id
import uos
import urandom
import hashlib
import ucryptolib
import json
import gc
import sys
import io

aZ09 = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'


def load_key(p_key=False):
    id = hexlify(unique_id()).decode()
    buff_key = b''
    with open('upy_pub_rsa{}.key'.format(id), 'rb') as keyfile:
        while True:
            try:
                buff = keyfile.read(2000)
                if buff != b'':
                    buff_key += buff
                else:
                    break
            except Exception as e:
                print(e)
    if p_key:
        print(buff_key)
    return buff_key


def upy_keygen(rsa_key, token, save_pass=True):
    if sys.platform == 'esp8266':
        raw_key_list = [line.encode() for line in rsa_key.decode().split('\n')[1:-2]]
    else:
        raw_key_list = [line for line in rsa_key.splitlines()[1:-1]]
    raw_key = b''
    for line in raw_key_list:
        raw_key += line
    random_token = token[:32]
    for b in random_token:
        raw_key += bytes(chr(raw_key[b]), 'utf-8')
    key_hash = hashlib.sha256()
    key_hash.update(raw_key)
    hashed_key = key_hash.digest()
    index_key = token[32:40]
    password_long = bytes([hashed_key[val] for val in index_key])
    password_short = bytes([aZ09[val % len(aZ09)] for val in password_long]).decode()

    if save_pass:
        with open('webrepl_cfg.py', 'wb') as wr_config:
            wr_config.write(bytes("PASS = '{}'\n".format(password_short), 'utf-8'))
    print('New password saved in the device!')
    return (password_short)


def upy_session_keygen(rsa_key, save_sessionkey=False, token=None):
    if sys.platform == 'esp8266':
        raw_key_list = [line.encode() for line in rsa_key.decode().split('\n')[1:-2]]
    else:
        raw_key_list = [line for line in rsa_key.splitlines()[1:-1]]
    raw_key = b''
    for line in raw_key_list:
        raw_key += line
    if token is None:
        random_token = uos.urandom(32) # send this
    else:
        random_token = token[:32]
    for b in random_token:
        raw_key += bytes(chr(raw_key[b]), 'utf-8')
    key_hash = hashlib.sha256()
    key_hash.update(raw_key)
    hashed_key = key_hash.digest()
    if token is None:
        if sys.platform == 'esp8266':
            index_pvkey = [urandom.getrandbits(4)+urandom.getrandbits(4) for i in range(32)]
        else:
            index_pvkey = [urandom.randrange(0, len(hashed_key)) for i in range(32)] # send this
    else:
        index_pvkey = token[32:64]
    pv_key = bytes([hashed_key[val] for val in index_pvkey])
    if token is None:
        if sys.platform == 'esp8266':
            index_ivkey = [urandom.getrandbits(4)+urandom.getrandbits(4) for i in range(16)]
        else:
            index_ivkey = [urandom.randrange(0, len(hashed_key)) for i in range(16)] # send this
    else:
        index_ivkey = token[64:]
    iv = bytes([hashed_key[val] for val in index_ivkey])

    if save_sessionkey:
        if token is None:
            with open('session.key', 'wb') as sess_config:
                sess_keys = dict(SKEY=pv_key, IV=iv)
                sess_config.write(json.dumps(sess_keys))
            print('New session.key saved in the device!')
        else:
            id = 'session{}.key'.format(hexlify(unique_id()).decode())
            with open(id, 'wb') as sess_config:
                sess_keys = dict(SKEY=pv_key, IV=iv)
                sess_config.write(json.dumps(sess_keys))
            print('New {} saved in the device!'.format(id))
    if token is None:
        return (pv_key, iv, [random_token + bytes(index_pvkey) + bytes(index_ivkey)])
    else:
        return (pv_key, iv)


class CRYPTOGRAPHER:
    def __init__(self, mode=2, key_enc=None, iv_enc=None, key_dec=None,
                 iv_dec=None, load_keys=False, buffer_size=2048):
        self.sess_keyfile_dev = 'session.key'
        self.sess_keyfile_host = 'session{}.key'.format(hexlify(unique_id()).decode())
        self.mode = mode
        self.block_len = 0
        self.key_e = key_enc
        self.iv_e = iv_enc
        self.msg_hex = b''
        self.buff = bytearray(buffer_size)
        self.buff_size = buffer_size
        self.err_buff = io.StringIO(100)
        self.buff_out = io.StringIO(500)
        self.gbls = globals()
        self.wrepl = None
        self.buff_crepl = None
        self.message_out = ''
        self.data_bytes = bytearray(20)
        self.rec_msg = ''
        self.resp_msg = ''
        if key_dec is None:
            self.key_d = key_enc
            self.iv_d = iv_enc
        else:
            self.key_d = key_dec
            self.iv_d = iv_dec
        if load_keys:
            try:
                with open(self.sess_keyfile_dev, 'r') as sess_config:
                    sess_keys_dev = json.load(sess_config)
                    self.key_e = sess_keys_dev['SKEY']
                    self.iv_e = sess_keys_dev['IV']

                with open(self.sess_keyfile_host, 'r') as sess_config:
                    sess_keys_host = json.load(sess_config)
                    self.key_d = sess_keys_host['SKEY']
                    self.iv_d = sess_keys_host['IV']
            except Exception as e:
                print('No session keys found in the device')
                pass

        self.enc = ucryptolib.aes(self.key_e, self.mode, self.iv_e)
        self.dec = ucryptolib.aes(self.key_d, self.mode, self.iv_d)

    def decrypt(self, msg):
        self.buff[:] = bytearray(self.buff_size)
        self.dec = ucryptolib.aes(self.key_d, self.mode, self.iv_d)
        self.dec.decrypt(msg, self.buff)
        gc.collect()
        return self.buff.decode().split('\x00')[0]

    def encrypt(self, msg):
        self.buff[:] = bytearray(self.buff_size)
        self.enc = ucryptolib.aes(self.key_e, self.mode, self.iv_e)
        self.data_bytes = msg.encode()
        self.block_len = len(self.data_bytes + b'\x00' * ((16 - (len(self.data_bytes) % 16)) % 16))
        self.enc.encrypt(self.data_bytes + b'\x00' * ((16 - (len(self.data_bytes) % 16)) % 16), self.buff)
        gc.collect()
        return bytes(self.buff[:self.block_len])

    def decrypt_hex(self, msg):
        self.msg_hex = unhexlify(msg)
        self.buff[:] = bytearray(self.buff_size)
        self.dec = ucryptolib.aes(self.key_d, self.mode, self.iv_d)
        self.dec.decrypt(self.msg_hex, self.buff)
        gc.collect()
        return self.buff.decode().split('\x00')[0]

    def encrypt_hex(self, msg):
        self.buff[:] = bytearray(self.buff_size)
        self.enc = ucryptolib.aes(self.key_e, self.mode, self.iv_e)
        self.data_bytes = msg.encode()
        self.block_len = len(self.data_bytes + b'\x00' * ((16 - (len(self.data_bytes) % 16)) % 16))
        self.enc.encrypt(self.data_bytes + b'\x00' * ((16 - (len(self.data_bytes) % 16)) % 16), self.buff)
        gc.collect()
        return hexlify(bytes(self.buff[:self.block_len]))

    def encrypt_hex_bytes(self, msg):
        self.buff[:] = bytearray(self.buff_size)
        self.enc = ucryptolib.aes(self.key_e, self.mode, self.iv_e)
        self.data_bytes = msg
        self.block_len = len(self.data_bytes + b'\x00' * ((16 - (len(self.data_bytes) % 16)) % 16))
        self.enc.encrypt(self.data_bytes + b'\x00' * ((16 - (len(self.data_bytes) % 16)) % 16), self.buff)
        gc.collect()
        return hexlify(bytes(self.buff[:self.block_len]))

    def recv_send(self, msg):
        rec_msg = self.decrypt_hex(msg)
        return self.encrypt_hex(rec_msg)

    def crepl(self, msg):
        self.rec_msg = self.decrypt_hex(msg)
        self.buff_out = io.StringIO(500)
        try:
            self.wrepl = uos.dupterm(self.buff_out, 0)
            self.resp_msg = eval(self.rec_msg)
            # self.buff_crepl = uos.dupterm(self.wrepl, 0)
            if self.resp_msg is None:
                self.message_out = self.buff_out.getvalue()
                self.buff_crepl = uos.dupterm(self.wrepl, 0)
                if self.message_out == '':
                    gc.collect()
                    return None
                else:
                    return self.encrypt_hex(self.message_out)
            else:
                self.buff_crepl = uos.dupterm(self.wrepl, 0)
                return self.encrypt_hex(str(self.resp_msg))
        except Exception as e:
            try:
                # self.buff_crepl = uos.dupterm(self.wrepl, 0)
                exec(self.rec_msg, self.gbls)
                self.message_out = self.buff_out.getvalue()
                self.buff_crepl = uos.dupterm(self.wrepl, 0)
                if self.message_out == '':
                    gc.collect()
                    return None
                else:
                    return self.encrypt_hex(self.message_out)
            except Exception as e:
                self.buff_crepl = uos.dupterm(self.wrepl, 0)
                sys.print_exception(e, self.err_buff)
                self.resp_msg = self.err_buff.getvalue()
                self.err_buff.seek(0)
                return self.encrypt_hex(self.resp_msg)

# TODO: ENCRYPT/DECRYPT FILE

# TODO: DECRYPT/ENCRYPT MESSAGE

# TODO: WCRYPL (WEB CRYTO REPL)



# µPing (MicroPing) for MicroPython
# copyright (c) 2018 Shawwwn <shawwwn1@gmail.com>
# License: MIT

# Internet Checksum Algorithm
# Author: Olav Morken
# https://github.com/olavmrk/python-ping/blob/master/ping.py
# @data: bytes

# ping statistics, loop mode and KeyboardInterrupt handler, + esp8266 compatible
# copyright (c) 2020 Carglglz
# License: MIT


def checksum(data):
    if len(data) & 0x1:  # Odd number of bytes
        data += b'\0'
    cs = 0
    for pos in range(0, len(data), 2):
        b1 = data[pos]
        b2 = data[pos + 1]
        cs += (b1 << 8) + b2
    while cs >= 0x10000:
        cs = (cs & 0xffff) + (cs >> 16)
    cs = ~cs & 0xffff
    return cs


def stddev(data):
    N = len(data)
    avg = sum(data)/N
    num = sum([(x-avg)**2 for x in data])
    den = N - 1
    stddev = (num/den)**0.5
    return stddev


def ping(host, count=4, timeout=5000, interval=10, quiet=False, size=64,
         rtn=True, loop=False, int_loop=800):
    import utime
    import uselect
    import uctypes
    import usocket
    import ustruct
    import urandom
    from sys import platform
    import gc
    from array import array

    # prepare packet
    assert size >= 16, "pkt size too small"
    pkt = b'Q'*size
    pkt_desc = {
        "type": uctypes.UINT8 | 0,
        "code": uctypes.UINT8 | 1,
        "checksum": uctypes.UINT16 | 2,
        "id": uctypes.UINT16 | 4,
        "seq": uctypes.INT16 | 6,
        "timestamp": uctypes.UINT64 | 8,
    }  # packet header descriptor
    h = uctypes.struct(uctypes.addressof(pkt), pkt_desc, uctypes.BIG_ENDIAN)
    h.type = 8  # ICMP_ECHO_REQUEST
    h.code = 0
    h.checksum = 0
    if platform == 'esp8266':
        h.id = urandom.getrandbits(16)
    else:
        h.id = urandom.randint(0, 65535)
    h.seq = 1
    time_data = array("f", (0 for _ in range(0)))

    # init socket
    sock = usocket.socket(usocket.AF_INET, usocket.SOCK_RAW, 1)
    sock.setblocking(0)
    sock.settimeout(timeout/1000)
    addr = usocket.getaddrinfo(host, 1)[0][-1][0]  # ip address
    sock.connect((addr, 1))
    not quiet and print("PING %s (%s): %u data bytes" % (host, addr, len(pkt)))
    seq_loop = -1
    try:
        if loop:
            n_trans = 0
            n_recv = 0
            while True:
                gc.collect()
                utime.sleep_ms(int_loop)
                count = 1
                seq_loop += 1
                seqs = list(range(1, count+1))  # [1,2,...,count]
                c = 1
                t = 0
                finish = False
                while t < timeout:
                    if t == interval and c <= count:
                        # send packet
                        h.checksum = 0
                        h.seq = c
                        h.timestamp = utime.ticks_us()
                        h.checksum = checksum(pkt)
                        if sock.send(pkt) == size:
                            n_trans += 1
                            t = 0  # reset timeout
                        else:
                            seqs.remove(c)
                            if loop:
                                count += 1
                                seqs.append(count)
                        c += 1

                    # recv packet
                    while 1:
                        socks, _, _ = uselect.select([sock], [], [], 0)
                        if socks:
                            resp = socks[0].recv(4096)
                            resp_mv = memoryview(resp)
                            h2 = uctypes.struct(uctypes.addressof(
                                resp_mv[20:]), pkt_desc, uctypes.BIG_ENDIAN)
                            # TODO: validate checksum (optional)
                            seq = h2.seq
                            # 0: ICMP_ECHO_REPLY
                            if h2.type == 0 and h2.id == h.id and (seq in seqs):
                                t_elapsed = (utime.ticks_us()-h2.timestamp) / 1000
                                ttl = ustruct.unpack('!B', resp_mv[8:9])[0]  # time-to-live
                                n_recv += 1
                                not quiet and print("{} bytes from {}: icmp_seq={} ttl={} time={:.3f} ms".format(
                                    len(resp), addr, seq_loop, ttl, t_elapsed))
                                time_data.append(t_elapsed)
                                seqs.remove(seq)
                                if len(seqs) == 0:
                                    finish = True
                                    break
                        else:
                            break

                    if finish:
                        break

                    utime.sleep_ms(1)
                    t += 1

        else:
            seqs = list(range(1, count+1))  # [1,2,...,count]
            c = 1
            t = 0
            n_trans = 0
            n_recv = 0
            finish = False
            while t < timeout:
                if t == interval and c <= count:
                    # send packet
                    h.checksum = 0
                    h.seq = c
                    h.timestamp = utime.ticks_us()
                    h.checksum = checksum(pkt)
                    if sock.send(pkt) == size:
                        n_trans += 1
                        t = 0  # reset timeout
                    else:
                        seqs.remove(c)
                        if loop:
                            count += 1
                            seqs.append(count)
                    c += 1

                # recv packet
                while 1:
                    socks, _, _ = uselect.select([sock], [], [], 0)
                    if socks:
                        resp = socks[0].recv(4096)
                        resp_mv = memoryview(resp)
                        h2 = uctypes.struct(uctypes.addressof(
                            resp_mv[20:]), pkt_desc, uctypes.BIG_ENDIAN)
                        # TODO: validate checksum (optional)
                        seq = h2.seq
                        # 0: ICMP_ECHO_REPLY
                        if h2.type == 0 and h2.id == h.id and (seq in seqs):
                            t_elapsed = (utime.ticks_us()-h2.timestamp) / 1000
                            ttl = ustruct.unpack('!B', resp_mv[8:9])[0]  # time-to-live
                            n_recv += 1
                            not quiet and print("{} bytes from {}: icmp_seq={} ttl={} time={:.3f} ms".format(
                                len(resp), addr, seq, ttl, t_elapsed))
                            time_data.append(t_elapsed)
                            seqs.remove(seq)
                            if loop:
                                count += 1
                                seqs.append(count)
                                utime.sleep_ms(int_loop)
                            if len(seqs) == 0:
                                finish = True
                                break
                    else:
                        break

                if finish:
                    if not loop:
                        break

                utime.sleep_ms(1)
                t += 1
        sock.close()
        if not quiet:
            print('--- {} ping statistics ---'.format(host))
            print("{} packets transmitted, {} packets received, {:.1f}% packet loss".format(
                n_trans, n_recv, (1-(n_recv/n_trans))*100))
            print("round-trip min/avg/max/stddev = {:.2f}/{:.2f}/{:.2f}/{:.2f} ms".format(min(time_data),sum(time_data)/len(time_data),max(time_data), stddev(time_data)))
        gc.collect()
        if rtn:
            return (n_trans, n_recv)
    except KeyboardInterrupt:
        # close
        sock.close()
        gc.collect()
        if not quiet:
            print('^C')
            print('--- {} ping statistics ---'.format(host))
            print("{} packets transmitted, {} packets received, {:.1f}% packet loss".format(
                n_trans, n_recv, (1-(n_recv/n_trans))*100))
            print("round-trip min/avg/max/stddev = {:.2f}/{:.2f}/{:.2f}/{:.2f} ms".format(min(time_data),sum(time_data)/len(time_data),max(time_data), stddev(time_data)))
        if rtn:
            return (n_trans, n_recv)
    except Exception as e:
        print(e)



import socket
import ssl
import os
import time
import sys
# import select
import gc
from ubinascii import hexlify
from machine import unique_id
from hashlib import sha256


class SSL_socket_client_repl:
    """
    SSL Socket client simple class repl
    """

    def __init__(self, host, port=8443, buff=1024, init=True, auth=True):
        self.cli_soc = None
        self.host = host
        self.port = port
        self.auth = auth
        self._key = 'SSL_key{}.der'.format(hexlify(unique_id()).decode())
        self._cert = 'SSL_certificate{}.der'.format(hexlify(unique_id()).decode())
        self.key = None
        self.cert = None
        with open(self._key, 'rb') as keyfile:
            self.key = keyfile.read()
        with open(self._cert, 'rb') as certfile:
            self.cert = certfile.read()
        self.cli_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # self.cli_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.wrepl = None
        if hasattr(os, 'dupterm_notify'):
            self.cli_soc.setsockopt(socket.SOL_SOCKET, 20, os.dupterm_notify)
        self.addr = socket.getaddrinfo(self.host, self.port)[0][-1]
        self.buff = bytearray(buff)
        self.buff_size = buff
        if init:
            print('>>> ')
            time.sleep(0.5)
            self.connect_SOC()
            # self.key = None
            # self.cert = None

    def connect_SOC(self):
        self.cli_soc.connect(self.addr)
        # self.cli_soc.settimeout(1)
        if self.auth:
            self.cli_soc = ssl.wrap_socket(self.cli_soc, key=self.key,
                                           cert=self.cert)
        # self.cli_soc = ssl.wrap_socket(self.cli_soc)
        else:
            self.cli_soc = ssl.wrap_socket(self.cli_soc)

        self.cli_soc.setblocking(False)
        # self.cli_soc.settimeout(1)
        print('>>> ')
        self.wrepl = os.dupterm(self.cli_soc, 0)

    def flush(self):
        flushed = 0
        while flushed == 0:
            try:
                self.cli_soc.recv(128)
            except Exception as e:
                flushed = 1
                print('flushed!')

    def send_message(self, message):
        self.cli_soc.write(bytes(message, 'utf-8'))

    def recv_message(self):
        try:
            self.buff[:] = self.cli_soc.read(self.buff_size)
            print(self.buff.decode())
        except Exception as e:
            pass

    def switch_wrepl(self):
        print('>>> ')
        self.cli_soc = os.dupterm(self.wrepl, 0)
        # print('WebREPL enabled!')

    def switch_ssl_repl(self):
        print('>>> ')
        self.wrepl = os.dupterm(self.cli_soc, 0)
        # print('SSL_REPL enabled!')


class SSL_socket_client_tool:
    """
    Socket client simple class io tool
    """

    def __init__(self, host, port=8433, init=False):
        self.cli_soc = None
        self.host = host
        self.port = port
        self.cli_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.cli_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.addr = socket.getaddrinfo(self.host, self.port)[0][-1]
        self.buff = bytearray(4096)
        self.readable = None
        self.writable = None
        self.exceptional = None

    def connect_SOC(self, key, cert):
        print('>>> ')
        # self.cli_soc.close()
        self.cli_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.cli_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        time.sleep(0.2)
        self.cli_soc.connect(self.addr)
        # self.cli_soc.settimeout(2)
        self.cli_soc = ssl.wrap_socket(self.cli_soc, key=key,
                                       cert=cert)
        self.cli_soc.setblocking(False)
        # self.cli_soc.settimeout(2)

    def put(self, filetoget, size, chunk=4096, dbg=False):
        local_size = 0
        chunk = chunk
        # put_soc = [self.cli_soc]
        # chunk_sizes = [chunk]
        print(size)
        if size < chunk:
            chunk = size
            chunk_rest = size % chunk
        else:
            # chunk_sizes = [chunk for i in range(size//chunk)] + [size % chunk]
            chunk_rest = size % chunk
        with open(filetoget, 'wb') as log:
            pass
        time.sleep(0.1)
        while True:
            try:
                # self.readable, self.writable, self.exceptional = select.select(put_soc,
                #                                                 put_soc,
                #                                                 put_soc)
                # if len(self.readable) == 1:
                if local_size == (size // (chunk))*(chunk):
                    chunk = chunk_rest

                # self.buff = self.cli_soc.read(chunk)  # 4 KB
                # print(self.buff)
                self.buff = self.cli_soc.read(chunk)
                if self.buff is None:
                    if dbg:
                        print('ZERO BYTES')
                    pass
                else:
                    if self.buff != b'':
                        local_size += len(self.buff)
                        with open(filetoget, 'ab') as log:
                            log.write(self.buff)
                        if local_size >= size:
                            break
                    else:
                        break
            except Exception as e:
                if e == KeyboardInterrupt:
                    break
                else:
                    sys.print_exception(e)
        print(local_size)
        self.buff = b''
        gc.collect()

    def get_size(self, filetoget):
        return os.stat(filetoget)[6]

    def get(self, filetoget, chunk=4096, dbg=False):
        size = self.get_size(filetoget)
        remote_size = 0
        chunk = chunk
        write_b = 0
        write_buff = 0
        if size < chunk:
            chunk = size
            chunk_rest = size % chunk
        else:
            # chunk_sizes = [chunk for i in range(size//chunk)] + [size % chunk]
            chunk_rest = size % chunk
        # final_file = b''
        time.sleep(0.1)
        with open(filetoget, 'rb') as log:
            self.buff = log.read(chunk)
            while True:
                try:
                    # print(len(chunk))
                    if self.buff != b'':
                        # in python use 'i'
                        write_b = self.cli_soc.write(self.buff)
                        if isinstance(write_b, int):
                            write_buff += write_b
                            if dbg:
                                print(write_b)
                            if remote_size == (size // (chunk))*(chunk):
                                chunk = chunk_rest
                            # final_file += chunk
                            self.buff = log.read(chunk)
                            remote_size += len(self.buff)
                    else:
                        # print('END OF FILE')
                        # soc.sendall(b'EOF\x8a\xb1\x1a\xcb\x11')
                        # soc.close()
                        # gc.collect()

                        time.sleep(0.1)
                        print(write_buff)
                        break
                except Exception as e:
                    if e == KeyboardInterrupt:
                        break
                    else:
                        sys.print_exception(e)
            # self.buff = b''
            # gc.collect()


def start(host, port, auth=True):
    ssl_repl_serv = SSL_socket_client_repl(host, port=port, auth=auth)
    return ssl_repl_serv


def get_files_cwd():
    return [file for file in os.listdir() if not os.stat(file)[0] & 0x4000]


def get_files_re(reg):
    return [file for file in get_files_cwd() if reg in file]


def get_files_dir(filelist):
    return [file for file in get_files_cwd() if file in filelist]



import sys
import os
import gc


class LTREE:

    def __repr__(self):
        self.__call__()
        return ""

    def __call__(self, path=".", level=0, is_last=False, is_root=True,
                 carrier="    ", hidden=False):
        os.chdir(path)
        r_path = path
        path = "."
        if hidden:
            l = os.listdir(path)
        else:
            l = [f for f in os.listdir(path) if not f.startswith('.')]
        nf = len([file for file in l if not os.stat("%s/%s" % (path, file))[0] & 0x4000])
        nd = len(l) - nf
        ns_f, ns_d = 0, 0
        l.sort()
        if len(l) > 0:
            last_file = l[-1]
        else:
            last_file = ''
        for f in l:
            st = os.stat("%s/%s" % (path, f))
            if st[0] & 0x4000:  # stat.S_IFDIR
                print(self._treeindent(level, f, last_file, is_last=is_last, carrier=carrier) + "  %s <dir>" % f)
                os.chdir(f)
                level += 1
                lf = last_file == f
                if level > 1:
                    if lf:
                        carrier += "     "
                    else:
                        carrier += "    │"
                ns_f, ns_d = self.__call__(level=level, is_last=lf,
                                           is_root=False, carrier=carrier,
                                           hidden=hidden)
                if level > 1:
                    carrier = carrier[:-5]
                os.chdir('..')
                level += (-1)
                nf += ns_f
                nd += ns_d
            else:
                print(self._treeindent(level, f, last_file, is_last=is_last, carrier=carrier) + "  %s" % (f))
        if is_root:
            print('{} directories, {} files'.format(nd, nf))
            if r_path != ".":
                os.chdir('..')
        else:
            return (nf, nd)

    def _treeindent(self, lev, f, lastfile, is_last=False, carrier=None):
        if lev == 0:
            return ""
        else:
            if f != lastfile:
                return carrier + "    ├────"
            else:
                return carrier + "    └────"


class DISK_USAGE:

    def __repr__(self):
        self.__call__()
        return ""

    def __call__(self, path=".", dlev=0, max_dlev=0, hidden=False):
        if path != ".":
                if not os.stat(path)[0] & 0x4000:
                    print('{:9} {}'.format(self.print_filesys_info(os.stat(path)[6]), path))
                else:
                    if hidden:
                        resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path)}
                    else:
                        resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path) if not dir.startswith('.')}
                    for dir in resp.keys():

                        if not os.stat(dir)[0] & 0x4000:
                            print('{:9} {}'.format(self.print_filesys_info(resp[dir]), dir))

                        else:
                            if dlev < max_dlev:
                                dlev += 1
                                self.__call__(path=dir, dlev=dlev, max_dlev=max_dlev, hidden=hidden)
                                dlev += (-1)
                            else:
                                print('{:9} {} {}'.format(self.print_filesys_info(self.get_dir_size_recursive(dir)), dir, '<dir>'))
                            gc.collect()

        else:
            if hidden:
                resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path)}
            else:
                resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path) if not dir.startswith('.')}
            for dir in resp.keys():

                if not os.stat(dir)[0] & 0x4000:
                    print('{:9} {}'.format(self.print_filesys_info(resp[dir]), dir))

                else:
                    if dlev < max_dlev:
                        dlev += 1
                        self.__call__(path=dir, dlev=dlev, max_dlev=max_dlev, hidden=hidden)
                        dlev += (-1)
                    else:
                        print('{:9} {} {}'.format(self.print_filesys_info(self.get_dir_size_recursive(dir)), dir, '<dir>'))

                    gc.collect()

    def print_filesys_info(self, filesize):
        _kB = 1024
        if filesize < _kB:
            sizestr = str(filesize) + " by"
        elif filesize < _kB**2:
            sizestr = "%0.1f KB" % (filesize / _kB)
        elif filesize < _kB**3:
            sizestr = "%0.1f MB" % (filesize / _kB**2)
        else:
            sizestr = "%0.1f GB" % (filesize / _kB**3)
        return sizestr

    def get_dir_size_recursive(self, dir):
        return sum([os.stat(dir+'/'+f)[6] if not os.stat(dir+'/'+f)[0] & 0x4000 else self.get_dir_size_recursive(dir+'/'+f) for f in os.listdir(dir)])


tree = LTREE()
du = DISK_USAGE()



import sys
try:
    import ubinascii as binascii
except:
    import binascii
try:
    import uhashlib as hashlib
except:
    import hashlib

DEBUG = 0

def server_handshake(sock):
    clr = sock.makefile("rwb", 0)
    l = clr.readline()
    #sys.stdout.write(repr(l))

    webkey = None

    while 1:
        l = clr.readline()
        if not l:
            raise OSError("EOF in headers")
        if l == b"\r\n":
            break
    #    sys.stdout.write(l)
        h, v = [x.strip() for x in l.split(b":", 1)]
        if DEBUG:
            print((h, v))
        if h == b'Sec-WebSocket-Key':
            webkey = v

    if not webkey:
        raise OSError("Not a websocket request")

    if DEBUG:
        print("Sec-WebSocket-Key:", webkey, len(webkey))

    respkey = webkey + b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    respkey = hashlib.sha1(respkey).digest()
    respkey = binascii.b2a_base64(respkey)[:-1]

    resp = b"""\
HTTP/1.1 101 Switching Protocols\r
Upgrade: websocket\r
Connection: Upgrade\r
Sec-WebSocket-Accept: %s\r
\r
""" % respkey

    if DEBUG:
        print(resp)
    sock.send(resp)


# Very simplified client handshake, works for MicroPython's
# websocket server implementation, but probably not for other
# servers.
def client_handshake(sock):
    cl = sock.makefile("rwb", 0)
    cl.write(b"""\
GET / HTTP/1.1\r
Host: echo.websocket.org\r
Connection: Upgrade\r
Upgrade: websocket\r
Sec-WebSocket-Key: foo\r
\r
""")
    l = cl.readline()
#    print(l)
    while 1:
        l = cl.readline()
        if l == b"\r\n":
            break
#        sys.stdout.write(l)



#
# upip - Package manager for MicroPython
#
# Copyright (c) 2015-2018 Paul Sokolovsky
#
# Licensed under the MIT license.
#
import sys
import os
import errno
import json
import zlib
import time
import shlex
import subprocess
import shutil
# import upip_host_tarfile as tarfile
# gc.collect()


debug = False
index_urls = ["https://micropython.org/pi", "https://pypi.org/pypi"]
install_path = None
cleanup_files = []
gzdict_sz = 16 + 15

file_buf = bytearray(512)


class NotFoundError(Exception):
    pass


def op_split(path):
    if path == "":
        return ("", "")
    r = path.rsplit("/", 1)
    if len(r) == 1:
        return ("", path)
    head = r[0]
    if not head:
        head = "/"
    return (head, r[1])


def op_basename(path):
    return op_split(path)[1]

# Expects *file* name


def _makedirs(name, mode=0o777):
    ret = False
    s = ""
    comps = name.rstrip("/").split("/")[:-1]
    if comps[0] == "":
        s = "/"
    for c in comps:
        if s and s[-1] != "/":
            s += "/"
        s += c
        try:
            os.mkdir(s)
            ret = True
        except OSError as e:
            if e.args[0] != errno.EEXIST and e.args[0] != errno.EISDIR:
                raise e
            ret = False
    return ret


def save_file(fname, subf):
    global file_buf
    with open(fname, "wb") as outf:
        while True:
            sz = subf.readinto(file_buf)
            if not sz:
                break
            outf.write(file_buf, sz)


def install_tar(f, prefix):
    meta = {}
    for info in f:
        # print(info)
        fname = info.name
        try:
            fname = fname[fname.index("/") + 1:]
        except ValueError:
            fname = ""

        save = True
        for p in ("setup.", "PKG-INFO", "README"):
                #print(fname, p)
                if fname.startswith(p) or ".egg-info" in fname:
                    if fname.endswith("/requires.txt"):
                        meta["deps"] = f.extractfile(info).read()
                    save = False
                    if debug:
                        print("Skipping", fname)
                    break

        if save:
            outfname = prefix + fname
            # if info.type != tarfile.DIRTYPE:
            #     if debug:
            #         print("Extracting " + outfname)
            #     _makedirs(outfname)
            #     subf = f.extractfile(info)
            #     save_file(outfname, subf)
    return meta


def expandhome(s):
    if "~/" in s:
        h = os.getenv("HOME")
        s = s.replace("~/", h + "/")
    return s


import ssl
import socket
warn_ssl = True


def url_open(url):
    global warn_ssl

    if debug:
        print(url)

    proto, _, host, urlpath = url.split('/', 3)
    try:
        ai = socket.getaddrinfo(host, 443, 0, socket.SOCK_STREAM)
    except OSError as e:
        fatal("Unable to resolve %s (no Internet?)" % host, e)
    # print("Address infos:", ai)
    ai = ai[0]

    s = socket.socket(ai[0], ai[1], ai[2])
    try:
        # print("Connect address:", addr)
        s.connect(ai[-1])

        if proto == "https:":
            context = ssl.create_default_context()

            s = context.wrap_socket(s, server_hostname=host)
            if warn_ssl:
                print("Warning: %s SSL certificate is not validated" % host)
                warn_ssl = False

        # MicroPython rawsocket module supports file interface directly
        s.write("GET /{} HTTP/1.0\r\nHost: {}\r\n\r\n".format(urlpath, host).encode('utf-8'))
        l = s.recv()
        protover, status, msg = l.split(None, 2)
        if status != b"200":
            if status == b"404" or status == b"301":
                raise NotFoundError("Package not found")
            raise ValueError(status)
        server_metadata = l.split(b'\n')[:-1]
        json_data = l.split(b'\n')[-1]
        while 1:
            l = s.recv()
            if not l:
                # raise ValueError("Unexpected EOF in HTTP headers")
                time.sleep(0.2)
                break
            if l == b'\r\n':
                break
    except Exception as e:
        s.close()
        raise e

    # print(server_metadata)
    return json_data


def get_pkg_metadata(name):
    for url in index_urls:
        try:
            f = url_open("%s/%s/json" % (url, name))
        except NotFoundError:
            continue
        try:
            return json.loads(f)
        finally:
            pass
    raise NotFoundError("Package not found")


def fatal(msg, exc=None):
    print("Error:", msg)
    if exc and debug:
        raise exc
    sys.exit(1)


def install_pkg(pkg_spec, install_path, read_pkg_info=False):
    kw = ["setup.", "PKG-INFO", "README"]
    print('Collecting: {}'.format(pkg_spec))
    data = get_pkg_metadata(pkg_spec)
    # print(data)
    latest_ver = data["info"]["version"]
    packages = data["releases"][latest_ver]
    del data
    # gc.collect()
    assert len(packages) == 1
    package_url = packages[0]["url"]
    if read_pkg_info:
        print("Reading PKG-INFO %s %s from %s" % (pkg_spec, latest_ver, package_url))
    else:
        print("Installing %s %s from %s" % (pkg_spec, latest_ver, package_url))
    curl_cmd_str = "curl -O '{}'".format(package_url)
    curl_cmd = shlex.split(curl_cmd_str)
    try:
        proc = subprocess.call(curl_cmd)
        # print('Done!')
    except KeyboardInterrupt:
        print('Operation cancelled')
    try:
        tarfile = [f for f in os.listdir() if pkg_spec.replace('micropython-', '') in f and f.endswith('.tar.gz')][0]
        shutil.unpack_archive(tarfile)
        os.remove(tarfile)
    except Exception as e:
        print(e)
    try:
        dir_pkg = [f for f in os.listdir() if pkg_spec.replace('micropython-', '') in f and not f.endswith('.py')][0]
        pkg_egg_info_dir = [f for f in os.listdir(dir_pkg) if f.endswith('.egg-info')][0]
        with open(dir_pkg+"/"+pkg_egg_info_dir+'/PKG-INFO', 'r') as pkginfo:
            pkg_info = pkginfo.read()
            # print(pkg_info)
    except Exception as e:
        print(e)
    if read_pkg_info:
        print(pkg_info)
        shutil.rmtree(dir_pkg)
    else:
        pkg_req = None
        try:
            print('Package requirements: ')
            with open(dir_pkg+"/"+pkg_egg_info_dir+'/requires.txt', 'r') as pkgreq:
                pkg_req = pkgreq.read().split('\n')
                for pk in pkg_req:
                    if pk != '':
                        print(pk)
        except Exception as e:
            print('None')

        if pkg_req:
            for pk in pkg_req:
                if pk != '':
                    req_content = install_pkg(pk, ".")

        pkg_content = [f for f in os.listdir(dir_pkg) if not any([f.startswith(p) for p in kw]) and ".egg-info" not in f]
        # print(pkg_content)
        dir_lib = 'lib'
        if dir_lib not in os.listdir():
            os.mkdir(dir_lib)
        for pkg_file in pkg_content:
            if os.stat(dir_pkg+'/'+pkg_file)[0] & 0x4000:
                if pkg_file not in os.listdir(dir_lib):
                    os.mkdir('/'.join([dir_lib, pkg_file]))
                rcopytree(dir_pkg+'/'+pkg_file, '/'.join([dir_lib, pkg_file]))
        for pkg_file in pkg_content:
            if not os.stat(dir_pkg+'/'+pkg_file)[0] & 0x4000:
                shutil.copyfile(dir_pkg+'/'+pkg_file, dir_lib+'/'+pkg_file)

        shutil.rmtree(dir_pkg)
        print('Successfully installed {}'.format(dir_pkg))
        return pkg_content, dir_pkg


def rcopytree(src, dst):
    for pkg_file in os.listdir(src):
        if os.stat(src+'/'+pkg_file)[0] & 0x4000:
            if pkg_file in os.listdir(dst):
                pass
            else:
                # print('Dest dir: ', dst+'/'+pkg_file)
                os.mkdir(dst+'/'+pkg_file)
            rcopytree(src+'/'+pkg_file, '/'.join([dst, pkg_file]))

    for pkg_file in os.listdir(src):
        if not os.stat(src+'/'+pkg_file)[0] & 0x4000:
            # print('Source file: ', src+'/'+pkg_file)
            shutil.copyfile(src+'/'+pkg_file, dst+'/'+pkg_file)


def install(to_install, install_path=None):
    # Calculate gzip dictionary size to use
    global gzdict_sz
    sz = 65538
    if sz <= 65536:
        gzdict_sz = 16 + 12

    if install_path is None:
        install_path = get_install_path()
    if install_path[-1] != "/":
        install_path += "/"
    if not isinstance(to_install, list):
        to_install = [to_install]
    print("Installing to: " + install_path)
    # sets would be perfect here, but don't depend on them
    installed = []
    try:
        while to_install:
            if debug:
                print("Queue:", to_install)
            pkg_spec = to_install.pop(0)
            if pkg_spec in installed:
                continue
            meta = install_pkg(pkg_spec, install_path)
            installed.append(pkg_spec)
            if debug:
                print(meta)
            deps = meta.get("deps", "").rstrip()
            if deps:
                deps = deps.decode("utf-8").split("\n")
                to_install.extend(deps)
    except Exception as e:
        print("Error installing '{}': {}, packages may be partially installed".format(
                pkg_spec, e),
              file=sys.stderr)


def get_install_path():
    global install_path
    if install_path is None:
        # sys.path[0] is current module's path
        install_path = sys.path[1]
    install_path = expandhome(install_path)
    return install_path


def cleanup():
    for fname in cleanup_files:
        try:
            os.unlink(fname)
        except OSError:
            print("Warning: Cannot delete " + fname)


def help():
    print("""\
upip - Simple PyPI package manager for MicroPython
Usage: micropython -m upip install [-p <path>] <package>... | -r <requirements.txt>
import upip; upip.install(package_or_list, [<path>])

If <path> is not given, packages will be installed into sys.path[1]
(can be set from MICROPYPATH environment variable, if current system
supports that).""")
    print("Current value of sys.path[1]:", sys.path[1])
    print("""\

Note: only MicroPython packages (usually, named micropython-*) are supported
for installation, upip does not support arbitrary code in setup.py.
""")


def main():
    global debug
    global index_urls
    global install_path
    install_path = None

    if len(sys.argv) < 2 or sys.argv[1] == "-h" or sys.argv[1] == "--help":
        help()
        return

    if sys.argv[1] != "install":
        fatal("Only 'install' command supported")

    to_install = []

    i = 2
    while i < len(sys.argv) and sys.argv[i][0] == "-":
        opt = sys.argv[i]
        i += 1
        if opt == "-h" or opt == "--help":
            help()
            return
        elif opt == "-p":
            install_path = sys.argv[i]
            i += 1
        elif opt == "-r":
            list_file = sys.argv[i]
            i += 1
            with open(list_file) as f:
                while True:
                    l = f.readline()
                    if not l:
                        break
                    if l[0] == "#":
                        continue
                    to_install.append(l.rstrip())
        elif opt == "-i":
            index_urls = [sys.argv[i]]
            i += 1
        elif opt == "--debug":
            debug = True
        else:
            fatal("Unknown/unsupported option: " + opt)

    to_install.extend(sys.argv[i:])
    if not to_install:
        help()
        return

    install(to_install)

    if not debug:
        cleanup()


if __name__ == "__main__":
    main()



import sys
try:
    import ubinascii as binascii
except:
    import binascii
try:
    import uhashlib as hashlib
except:
    import hashlib

DEBUG = 0


def server_handshake(sock, ssl=False):
    if not ssl:
        clr = sock.makefile("rwb", 0)
    else:
        clr = sock
    l = clr.readline()
    #sys.stdout.write(repr(l))

    webkey = None

    while 1:
        l = clr.readline()
        if not l:
            raise OSError("EOF in headers")
        if l == b"\r\n":
            break
    #    sys.stdout.write(l)
        h, v = [x.strip() for x in l.split(b":", 1)]
        if DEBUG:
            print((h, v))
        if h == b'Sec-WebSocket-Key':
            webkey = v

    if not webkey:
        raise OSError("Not a websocket request")

    if DEBUG:
        print("Sec-WebSocket-Key:", webkey, len(webkey))

    d = hashlib.sha1(webkey)
    d.update(b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11")
    respkey = d.digest()
    respkey = binascii.b2a_base64(respkey)[:-1]
    if DEBUG:
        print("respkey:", respkey)

    sock.write(b"""\
HTTP/1.1 101 Switching Protocols\r
Upgrade: websocket\r
Connection: Upgrade\r
Sec-WebSocket-Accept: """)
    sock.write(respkey)
    sock.write("\r\n\r\n")


# Very simplified client handshake, works for MicroPython's
# websocket server implementation, but probably not for other
# servers.
def client_handshake(sock, ssl=False):
    if not ssl:
        cl = sock.makefile("rwb", 0)
    else:
        cl = sock
    cl.write(b"""\
GET / HTTP/1.1\r
Host: echo.websocket.org\r
Connection: Upgrade\r
Upgrade: websocket\r
Sec-WebSocket-Key: foo\r
\r
""")
    if not ssl:
        l = cl.readline()
    #    print(l)
        while 1:
            l = cl.readline()
            if l == b"\r\n":
                break
    else:
        line = b''
        while b'\n' not in line:
            line += cl.recv(1)
    #    print(l)

        while 1:
            line = b''
            while b'\n' not in line:
                line += cl.recv(1)
            if line == b"\r\n":
                break
#        sys.stdout.write(l)



# This module should be imported from REPL, not run from command line.
import socket
import uos
import network
import uwebsocket
import websocket_helper
import wss_helper
import _webrepl
import ssl
from binascii import hexlify
from machine import unique_id


key = None
cert = None
listen_s = None
client_s = None
websslrepl = False


def setup_conn(port, accept_handler):
    global listen_s, websslrepl
    listen_s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    listen_s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    ai = socket.getaddrinfo("0.0.0.0", port)
    addr = ai[0][4]

    listen_s.bind(addr)
    listen_s.listen(1)
    if accept_handler:
        listen_s.setsockopt(socket.SOL_SOCKET, 20, accept_handler)
    for i in (network.AP_IF, network.STA_IF):
        iface = network.WLAN(i)
        if iface.active():
            if websslrepl:
                print("WebSecureREPL daemon started on wss://%s:%d" % (iface.ifconfig()[0], port))
            else:
                print("WebREPL daemon started on ws://%s:%d" % (iface.ifconfig()[0], port))
    return listen_s


def accept_conn(listen_sock):
    global client_s, key, cert, websslrepl
    cl, remote_addr = listen_sock.accept()
    prev = uos.dupterm(None)
    uos.dupterm(prev)
    if prev:
        print("\nConcurrent WebREPL connection from", remote_addr, "rejected")
        cl.close()
        return
    print("\nWebREPL connection from:", remote_addr)
    client_s = cl
    if websslrepl:
        if hasattr(uos, 'dupterm_notify'):
            cl.setsockopt(socket.SOL_SOCKET, 20, uos.dupterm_notify)
        cl = ssl.wrap_socket(cl, server_side=True, key=key, cert=cert)
        wss_helper.server_handshake(cl, ssl=True)
    else:
        websocket_helper.server_handshake(cl)
    ws = uwebsocket.websocket(cl, True)
    ws = _webrepl._webrepl(ws)
    cl.setblocking(False)
    # notify REPL on socket incoming data (ESP32/ESP8266-only)
    if not websslrepl:
        if hasattr(uos, 'dupterm_notify'):
            cl.setsockopt(socket.SOL_SOCKET, 20, uos.dupterm_notify)
    uos.dupterm(ws)


def stop():
    global listen_s, client_s
    uos.dupterm(None)
    if client_s:
        client_s.close()
    if listen_s:
        listen_s.close()


def start(port=8266, password=None, ssl=False):
    global key, cert, websslrepl
    if ssl:
        websslrepl = True
        port = 8833
        _key = 'SSL_key{}.der'.format(hexlify(unique_id()).decode())
        _cert = 'SSL_certificate{}.der'.format(hexlify(unique_id()).decode())
        try:
            with open(_key, 'rb') as keyfile:
                key = keyfile.read()
            with open(_cert, 'rb') as certfile:
                cert = certfile.read()
        except Exception as e:
            print('No key or certificate found')
    else:
        websslrepl = False
    stop()
    if password is None:
        try:
            import webrepl_cfg
            _webrepl.password(webrepl_cfg.PASS)
            setup_conn(port, accept_conn)
            print("Started webrepl in normal mode")
        except:
            print("WebREPL is not configured, run 'import webrepl_setup'")
    else:
        _webrepl.password(password)
        setup_conn(port, accept_conn)
        print("Started webrepl in manual override mode")


def start_foreground(port=8266):
    stop()
    s = setup_conn(port, None)
    accept_conn(s)


def set_ssl(flag):
    with open('ssl_flag.py', 'wb') as sslconfig:
        sslconfig.write(b'SSL = {}'.format(flag))
